﻿/*	OpenCLSGEMM.c
	This XOP aims at providing Igor Pro 7x64 or above with access to OpenCL SGEMM, in an effort to vastly speed up post-processing of data acquisition.
*/

#include "XOPStandardHeaders.h"			// Include ANSI headers, Mac headers, IgorXOP.h, XOP.h and XOPSupport.h
#include "OpenCLSGEMM.h"

// Global Variables
static int gCallSpinProcess = 1;		// Set to 1 to all user abort (cmd dot) and background processing.
static int
AddCStringToHandle(						// Concatenates C string to handle.
	const char *theStr,
	Handle theHand)
{
	return WMPtrAndHand(theStr, theHand, strlen(theStr));
}

#pragma pack(2)		// All structures passed to Igor are two-byte aligned
typedef struct GPU_SGEMMParams {	//For a test function to check the I/O between Igor Pro and C
	waveHndl	outputMatrix;	//waveHandle for the output matrix
	waveHndl	inputMatrix2;	//waveHandle for the second input matrix
	waveHndl	inputMatrix1;	//waveHandle for the first input matrix
	double		result;			//error reporting for the calculation
} GPU_SGEMMParams;
#pragma pack()		// Reset structure alignment to default.

int GPU_SGEMM(GPU_SGEMMParams *p) {

	char	buf[256];
	float	*inM1, *inM2, *outM;
	//char	waveName1[MAX_OBJ_NAME + 1];
	//char	waveName2[MAX_OBJ_NAME + 1];
	//int		waveType1, waveType2;
	//int		dimension1, dimension2, dimension3
	int		numDimensions1, numDimensions2, numDimensions3;
	CountInt dimensionSizes1[MAX_DIMENSIONS + 1];
	CountInt dimensionSizes2[MAX_DIMENSIONS + 1];
	CountInt dimensionSizes3[MAX_DIMENSIONS + 1];
	//int		m, n, k;	//Iteratives for matrix computation with nested for loops
	int		M, N, K;	//Matrices dimension related constants, M is number of elements in a row of inV1, N for number of elements in a col of inV2,
						//K is number of col in inV1 == number of row inV2
	//float	acc;		//accumulator for nested matrix loops
	//char	dimensionUnits[MAX_DIMENSIONS][MAX_UNIT_CHARS + 1];
	//IndexInt element;
	//char	dataUnits[MAX_UNIT_CHARS + 1];
	//double	dataFullScaleMax, dataFullScaleMin;
	//double	sfA[MAX_DIMENSIONS];
	//double	sfB[MAX_DIMENSIONS];
	//char	dimLabel[MAX_DIM_LABEL_BYTES + 1];
	int		result;

	// Processing waves input from Igor Pro
	// Check if waves exist
	if (p->inputMatrix1 == NIL || p->inputMatrix2 == NIL || p->outputMatrix == NIL) {
		p->result = -1;					// Tell Igor that function return value is undefined.
		return NON_EXISTENT_WAVE;			// Make sure wave exists.
	}

	// Check if waves are single precision
	if (!(WaveType(p->inputMatrix1) == NT_FP32) || !(WaveType(p->inputMatrix2) == NT_FP32) || !(WaveType(p->outputMatrix) == NT_FP32)) {
		p->result = -1;
		return WAVE_DType_MISMATCH;			// Make sure waves are single precision floating points.
	}

	// Get number of used dimensions in inputMatrix1.
	if (result = MDGetWaveDimensions(p->inputMatrix1, &numDimensions1, dimensionSizes1))
		return result;
	// Get number of used dimensions in inputMatrix2.
	if (result = MDGetWaveDimensions(p->inputMatrix2, &numDimensions2, dimensionSizes2))
		return result;
	// Get number of used dimensions in inputMatrix2.
	if (result = MDGetWaveDimensions(p->outputMatrix, &numDimensions3, dimensionSizes3))
		return result;

	// Make sure matrices have the correct dimension for matrix multiplication, where inM1*inM2=outM
	if (dimensionSizes1[1] != dimensionSizes2[0] || dimensionSizes1[0] != dimensionSizes3[0] || dimensionSizes2[1] != dimensionSizes3[1]) {
		p->result = -1;
		return WAVE_DIM_MISMATCH;			// Make sure dimensions of waves match.
	}

	// Assigning dimensions to M, N, K
	M = dimensionSizes1[0];	//Number of rows of left matrix (inM1)
	N = dimensionSizes2[1]; //Number of col of right matrix (inM2)
	K = dimensionSizes1[1]; //Number of col of inM1 == Number of row of inM2

	// Get pointers to the wave data to store the samples in
	inM1 = (float*)WaveData(p->inputMatrix1);
	inM2 = (float*)WaveData(p->inputMatrix2);
	outM = (float*)WaveData(p->outputMatrix);

	// Calculate host memory allocated for inM1, inM2 and outM
	unsigned int size_inM1 = dimensionSizes1[0] * dimensionSizes1[1];
	unsigned int mem_size_inM1 = sizeof(float) * size_inM1;
	unsigned int size_inM2 = dimensionSizes2[0] * dimensionSizes2[1];
	unsigned int mem_size_inM2 = sizeof(float) * size_inM2;
	unsigned int size_outM = dimensionSizes1[0] * dimensionSizes2[1];
	unsigned int mem_size_outM = sizeof(float) * size_outM;

	// Prepairing the host for talking to GPU, obtaining platform and device resource for running kernel
	// Working only with a single GPU for now
	// credit to http://opencl.codeplex.com/ and http://www.es.ele.tue.nl/~mwijtvliet/5KK73/?page=mmopencl
	// credit to https://stackoverflow.com/questions/24200539/cant-get-my-graphics-card-device-in-opencl for ideas on searching over platforms
	cl_int error = 0;   // Used to handle error codes
	cl_uint p_count, d_count;
	cl_context context;
	cl_command_queue queue;
	cl_device_id device = 0;

	// Count the number of platforms
	if (clGetPlatformIDs(0, NULL, &p_count) != CL_SUCCESS)
	{
		p->result = -1;
		return NO_PLATFORM;
	}
	cl_platform_id platform_ids[100];
	clGetPlatformIDs(p_count, platform_ids, NULL);


	// Search all platforms for any GPU devices
	bool found = false;
	for (int i = 0; i<static_cast<int>(p_count); i++)
		if (clGetDeviceIDs(platform_ids[i], CL_DEVICE_TYPE_GPU, 1, &device, &d_count) == CL_SUCCESS) {
			found = true;
			break;
		}
	if (!found) {
		p->result = error;
		return NO_GPU;
	}

	// Create a compute context 
	context = clCreateContext(0, 1, &device, NULL, NULL, &error);
	if (!context) {
		p->result = error;
		return CREATE_CONTEXT_FAIL;
	}

	// Create a command commands
	queue = clCreateCommandQueueWithProperties(context, device, 0, &error);
	if (!queue) {
		p->result = error;
		return CREATE_QUEUE_FAIL;
	}

	// Building the kernal from OpenCLSGEMM.cl
	/* create a CL program using the kernel source */
	// credit to AMD SDK 3.0 samples, MatrixMulDouble + http://www.es.ele.tue.nl/~mwijtvliet/5KK73/?page=mmopencl
	SDKFile kernelFile;
	std::string kernelStr = std::string("OpenCLSGEMM.cl");
	//std::string kernelPath = getPath();
	//kernelPath.append(kernelStr.c_str());
	//std::string kernelPath = "D:\\kernels\\OpenCLSGEMM.cl";	
	//Using absolute path for convenience. You can comment this out, and use the previous two lines instead.
	std::string kernelPath = "C:\\Program Files\\WaveMetrics\\Igor Pro 7 Folder\\Igor Extensions (64-bit)\\OpenCLSGEMM.cl";
	if (!kernelFile.open(kernelPath.c_str())) {
		p->result = error;
		return LOAD_KERNEL_FAIL;
	}
	const char *source = kernelFile.source().c_str();
	size_t sourceSize[] = { strlen(source) };
	cl_program program = clCreateProgramWithSource(
		context,
		1,
		&source,
		sourceSize,
		&error);
	if (error != CL_SUCCESS) {
		p->result = error;
		return CREATE_PROGRAM_FAIL;
	}

	//program = clCreateProgramWithSource(context, 1, (const char **)& KernelSource, NULL, &err);
	//if (!program)
	//{
	//	printf("Error: Failed to create compute program!\n");
	//	return EXIT_FAILURE;
	//}

	// Build the program executable
	error = clBuildProgram(program, 1, &device, NULL, NULL, NULL);
	if (error != CL_SUCCESS) {
		p->result = error;
		return CREATE_EXE_FAIL;
		//For now I skip the detailed error reporting.
		//size_t len;
		//char buffer[2048];
		//printf("Error: Failed to build program executable!\n");
		//clGetProgramBuildInfo(program, device_id, CL_PROGRAM_BUILD_LOG, sizeof(buffer),
		//	buffer, &len);

		//printf("%s\n", buffer);
		//exit(1);
	}

	// Create the compute kernel in the program we wish to run
	cl_kernel kernel = clCreateKernel(program, "OpenCLSGEMM", &error);
	if (!kernel || error != CL_SUCCESS) {
		p->result = error;
		return CREATE_COM_KERNEL_FAIL;
	}

	// Check if the total matrix sizes exceeded the maximum memory size of the GPU
	cl_ulong global_mem_size;
	clGetDeviceInfo(device, CL_DEVICE_GLOBAL_MEM_SIZE, sizeof(global_mem_size), &global_mem_size, NULL);
	int mem_exceeded = static_cast<int>(global_mem_size*0.95) - static_cast<int>(mem_size_inM1 + mem_size_inM2 + mem_size_outM);

	if (mem_exceeded < 0) {
		p->result = -1;
		return GPU_MEM_EXCEEDED;
	}

	// Prepare OpenCL memory objects
	cl_mem d_inM1 = clCreateBuffer(context, CL_MEM_READ_ONLY, mem_size_inM1, NULL, NULL);
	cl_mem d_inM2 = clCreateBuffer(context, CL_MEM_READ_ONLY, mem_size_inM2, NULL, NULL);
	cl_mem d_outM = clCreateBuffer(context, CL_MEM_READ_WRITE, mem_size_outM, NULL, NULL);

	// Copy matrices to the GPU
	clEnqueueWriteBuffer(queue, d_inM1, CL_TRUE, 0, mem_size_inM1, inM1, 0, NULL, NULL);
	clEnqueueWriteBuffer(queue, d_inM2, CL_TRUE, 0, mem_size_inM2, inM2, 0, NULL, NULL);
	clEnqueueWriteBuffer(queue, d_outM, CL_TRUE, 0, mem_size_outM, outM, 0, NULL, NULL);

	// Preparing the kernel for execution
	error = clSetKernelArg(kernel, 0, sizeof(int), (void*)&M);
	error = clSetKernelArg(kernel, 1, sizeof(int), (void*)&N);
	error = clSetKernelArg(kernel, 2, sizeof(int), (void*)&K);
	error = clSetKernelArg(kernel, 2, sizeof(int), (void*)&K);
	error = clSetKernelArg(kernel, 3, sizeof(cl_mem), (void*)&d_inM1);
	error = clSetKernelArg(kernel, 4, sizeof(cl_mem), (void*)&d_inM2);
	error = clSetKernelArg(kernel, 5, sizeof(cl_mem), (void*)&d_outM);

	const size_t TS = 16;
	const size_t globalRow = static_cast<const size_t>(M), globalCol = static_cast<const size_t>(N);
	const size_t local[2] = { TS, TS/WPT };
	const size_t global[2] = { globalRow, globalCol/WPT };

	// Executing the kernel for the matrix multiplication
	cl_event event;
	error = clEnqueueNDRangeKernel(queue, kernel, 2, NULL,
		global, local, 0, NULL, &event);
	error = clWaitForEvents(1, &event);

	if (error != CL_SUCCESS)
	{
		p->result = -1;
		return COMPUTE_FAIL;
	}

	// Retrieve results from the GPU
	error = clEnqueueReadBuffer(queue, d_outM, CL_TRUE, 0, mem_size_outM, outM, 0, NULL, NULL);

	if (error != CL_SUCCESS)
	{
		p->result = -1;
		return MEM_CP_DTH_FAIL;
	}

	// Clean up GPU resource after finishing
	// Free the OpenCL memory objects
	clReleaseMemObject(d_inM1);
	clReleaseMemObject(d_inM2);
	clReleaseMemObject(d_outM);

	// Clean-up OpenCL 
	clReleaseCommandQueue(queue);
	clReleaseContext(context);
	clReleaseProgram(program);
	clReleaseKernel(kernel);

	p->result = error;
	return 0;
}

/*	RegisterFunction()
	
	Igor calls this at startup time to find the address of the
	XFUNCs added by this XOP. See XOP manual regarding "Direct XFUNCs".
*/
static XOPIORecResult
RegisterFunction()
{
	int funcIndex;

	funcIndex = (int)GetXOPItem(0);		// Which function is Igor asking about?
	switch (funcIndex) {
		case 0:						// WAGetWaveInfo(wave)
			return (XOPIORecResult)GPU_SGEMM;
			break;
		//case 3:						// WAFill3DWaveStorageMethod(wave)
		//	return (XOPIORecResult)WAFill3DWaveStorageMethod;
		//	break;
		//case 4:						// WAModifyTextWave(wave, prependString, appendString)
		//	return (XOPIORecResult)WAModifyTextWave;
		//	break;
	}
	return 0;
}

/*	DoFunction()
	
	Igor calls this when the user invokes one if the XOP's XFUNCs
	if we returned NIL for the XFUNC from RegisterFunction. In this
	XOP, we always use the direct XFUNC method, so Igor will never call
	this function. See XOP manual regarding "Direct XFUNCs".
*/
//static int
//DoFunction()
//{
//	int funcIndex;
//	void *p;				// Pointer to structure containing function parameters and result.
//	int err;
//
//	funcIndex = (int)GetXOPItem(0);	// Which function is being invoked ?
//	p = (void*)GetXOPItem(1);		// Get pointer to params/result.
//	switch (funcIndex) {
//		case 0:						// WAGetWaveInfo(wave)
//			err = WAGetWaveInfo((WAGetWaveInfoParams*)p);
//			break;
//		case 1:						// WAFill3DWaveDirectMethod(wave)
//			err = WAFill3DWaveDirectMethod((WAFill3DWaveDirectMethodParams*)p);
//			break;
//		case 2:						// WAFill3DWavePointMethod(wave)
//			err = WAFill3DWavePointMethod((WAFill3DWavePointMethodParams*)p);
//			break;
//		case 3:						// WAFill3DWaveStorageMethod(wave)
//			err = WAFill3DWaveStorageMethod((WAFill3DWaveStorageMethodParams*)p);
//			break;
//		case 4:						// WAModifyTextWave(wave, prependString, appendString)
//			err = WAModifyTextWave((WAModifyTextWaveParams*)p);
//			break;
//	}
//	return(err);
//}

/*	XOPEntry()

	This is the entry point from the host application to the XOP for all messages after the
	INIT message.
*/
extern "C" void
XOPEntry(void)
{	
	XOPIORecResult result = 0;

	switch (GetXOPMessage()) {
		//case FUNCTION:						// Our external function being invoked ?
		//	result = DoFunction();
		//	break;

		case FUNCADDRS:
			result = RegisterFunction();
			break;
	}
	SetXOPResult(result);
}

/*	XOPMain(ioRecHandle)

	This is the initial entry point at which the host application calls XOP.
	The message sent by the host must be INIT.
	
	XOPMain does any necessary initialization and then sets the XOPEntry field of the
	ioRecHandle to the address to be called for future messages.
*/
HOST_IMPORT int
XOPMain(IORecHandle ioRecHandle)		// The use of XOPMain rather than main means this XOP requires Igor Pro 6.20 or later
{	
	XOPInit(ioRecHandle);				// Do standard XOP initialization.
	SetXOPEntry(XOPEntry);				// Set entry point for future calls.
	
	if (igorVersion < 620) {			// Requires Igor Pro 6.20 or later.
		SetXOPResult(OLD_IGOR);			// OLD_IGOR is defined in OpenCLSGEMM.h and there are corresponding error strings in OpenCLSGEMM.r and OpenCLSGEMMWinCustom.rc.
		return EXIT_FAILURE;
	}

	SetXOPResult(0L);
	return EXIT_SUCCESS;
}
